# gasonline-analyser-v2.R

#
# Version 2
#
# Made at the direction of Whitney Newey
#
# Author:
# Luis Hoderlein, 2019
#

### LICENCE ###
# If you obtain this software, you can do whatever you want with it

# Authors note:
# 1. R automatically transposes vectors (and matrix subsections, which are vectors),
# to fit the operation performed on it.

# 2. Many of the functions included do NOT recompute values, instead pulling them from the global environment
# IF THE FUNCTIONS ARE NOT RUN IN SEQUENCE THERE WILL BE ERRORS, and if not, the numbers will be wrong
# you MUST call calcInfluenceFunction() before you can call calcEndogeneityEffect()
# you MUST call calcEndogeneityEffect() before you can call calcEndogeneityEffectVariance()
# the order supplied in the OUTPUT section of this program is correct
# this is done in order to significantly reduce code complexity, improve runtime, and to eliminate a potential source of bugs.

### CONSTANTS ###
e = 2.71828182845904523536028747135266249775724709369995

### DATA & OTHER INITIALIZATION ###
{
  options(expressions=100000) # overrides the maximum recursion depth
  
  raw_data = read.csv("Gasoline Data.csv") # read in data
  
  data = data.frame(tmp=rep(0,length(raw_data[[1]])))
  data$state_number = raw_data[[1]]
  data$ln_quant = raw_data[[2]]
  data$ln_price_dollars = raw_data[[3]]
  data$ln_income_dollars = raw_data[[4]]
  data$dist_gulf_mex = raw_data[[5]]
  data$num_drivers_household = e^raw_data[[6]] # this was in logs
  data$mean_age = raw_data[[8]]
  data$state_tax_percent = raw_data[[10]]

  # some data is not needed, and thus is excluded
  
  #data$dummy_public_transit = raw_data[[7]]
  #data$ln_price_cents = raw_data[[9]]
  
  # re-order the data
  data = subset(data, select=-tmp)
  data = data[order(data$state_number),]
  
  # compute budget shares
  data$budget_share = e^data$ln_price_dollars*e^data$ln_quant/e^data$ln_income_dollars*52 # quantity is weekly, while income is yearly
  
  # ignore bad budget shares (budgest shares > 1)
  initial_data_length = length(data$budget_share)
  data = data[-which(data$budget_share > 1),]
  
  # print the initialization notice
  print(":: INIT NOTICE :: gasoline-analyser-v2 ::")
  print(paste(sep="", "There was a ", (initial_data_length-length(data$budget_share)), " (",
              round(digits=5,(initial_data_length-length(data$budget_share))/length(data$budget_share)*100),
              "%) reduction in size after pruning budget shares larger than 1"))
  
  rm(initial_data_length) # clean
}

### SUPPORT FUNCTIONS ###

print = function(args) {
  writeLines(paste(sep="","  ",args))
}

exec = function(...) {
  eval(parse(text=paste(...,sep="")))
}

### FUNCTIONS ###

# this function calculate an influence function for a set of regressors
# note that price powers specifies the power of ln price contained in that regressor:
# e.g. (ln p)^2 * (ln y)^3 --> 2
# both should be passed in as lists
calcInfluenceFunction = function(regressors, price_powers) {
  if (length(regressors) != length(price_powers)) {
    print("Regressors & Price Powers do not match in length")
    return(NULL)
  }
  
  regressor_count <<- length(regressors)+1
  observation_count <<- length(regressors[[1]])

  print(paste("Calculating influence function with", regressor_count, "regressors and", observation_count, "observations"))
  
  # create a matrix representing the regresors
  x <<- matrix(nrow = observation_count, ncol = regressor_count)
  x[,1] <<- 1
  
  i = 1
  for (reg in regressors) {
    i = i + 1
    x[,i] <<- reg
  }
  
  p_bar = mean(e^data$ln_price_dollars) # the average price
  p_tilde = 1.1 * p_bar # 1.1 times the average price
  
  ln_p_star <<- vector(length = observation_count) # generated prices
  for (i in 1:observation_count) {
    ln_p_star[i] <<- log(runif(1, min=p_bar, max=p_tilde))
  }
  
  # create a matrix representing the regressors, but with the generated prices
  x_star <<- matrix(nrow = observation_count, ncol = regressor_count)
  x_star[,1] <<- 1
  
  i = 0
  for (reg in regressors) {
    i = i + 1
    x_star[,i+1] <<- reg * (ln_p_star ^ price_powers[[i]]) / (data$ln_price_dollars ^ price_powers[[i]])
  }
  
  # create the 'a' matrix
  a <<- x_star * (p_tilde - p_bar) * (e^data$ln_income_dollars / e^ln_p_star)
  
  # compute a_hat
  a_hat <<- vector(length = regressor_count)
  for (i in 1:regressor_count) {
    a_hat[i] <<- mean(a[,i])
  }
  
  sigma_hat <<- matrix(data=rep(0, regressor_count ^ 2), nrow = regressor_count, ncol = regressor_count)
  for (i in 1:observation_count) {
    sigma_hat <<- sigma_hat + t(t(x[i,])) %*% t(x[i,])
  }
  sigma_hat <<- sigma_hat / observation_count
  
  regr <<- lm(formula = data$budget_share ~ ., data = data.frame(regressors)) # run the regression, and keep it just in case
  beta_hat <<- unname(regr$coefficients)
  
  # compute influence function
  influence = vector(length = observation_count)
  for (i in 1:observation_count) {
    influence[i] = t(a[i,]) %*% beta_hat - t(a_hat) %*% beta_hat + 
      t(a_hat) %*% solve(sigma_hat) %*% x[i,] %*% (data$budget_share[[i]] - x[i,] %*% beta_hat)
  }
  
  return(influence)
}


# calculates the effect of endogeneity
# follows the same rules for arguments as before
calcEndogeneityEffect = function(regressors, price_powers) {
  if (length(regressors) != length(price_powers)) {
    print("Regressors & Price Powers do not match in length")
    return(NULL)
  }
  
  regressor_count <<- length(regressors)+1
  observation_count <<- length(regressors[[1]])
  
  print(paste("Calculating endogeneity effect with", regressor_count, "regressors and", observation_count, "observations"))
  
  # create a matrix representing the regressors, but adding instrumental variables
  z <<- matrix(nrow = observation_count, ncol = regressor_count)
  z[,1] <<- 1
  
  i = 0
  for (reg in regressors) {
    i = i + 1
    z[,i+1] <<- reg * (data$dist_gulf_mex ^ price_powers[[i]]) / (data$ln_price_dollars ^ price_powers[[i]])
  }
  
  # calculate delta_hat
  delta_hat <<- solve(t(z) %*% x) %*% (t(z) %*% data$budget_share)
  
  # calculate u_hat
  u_hat <<- data$budget_share - x %*% delta_hat
  
  # calculate & return the effect
  effect = a_hat %*% solve(sigma_hat) %*% t(x) %*% (u_hat / observation_count) 
  return(effect)
}

# calculate the variance of the effect of endogeneity
calcEndogeneityEffectVariance = function() {
  optim1 <<- solve(t(z) %*% x) # only compute (t(z) %*% x)^-1 only once, as the value is the same each iteration, to speed up the program
  e_hat <<- vector(length = observation_count)
  for (i in 1:observation_count) {
    e_hat[i] <<- data$budget_share[[i]] - t(x[i,]) %*% optim1 %*% t(z) %*% data$budget_share; 
  }
  
  # compute b_hat
  b_hat <<- rep(0,regressor_count)
  for (i in 1:observation_count) {
    b_hat <<- b_hat + x[i,] * e_hat[i]
  }
  b_hat <<- b_hat / observation_count
  
  sigma_hat_inverse = solve(sigma_hat)
  
  g_hat_transpose <<- t(a_hat) %*% sigma_hat_inverse # 1x5
  h_hat <<- sigma_hat_inverse %*% b_hat # 5x1
  
  # compute phi1_hat
  phi1_hat <<- vector(length = observation_count)
  for (i in 1:observation_count) {
    phi1_hat[i] <<- t(a[i,] - a_hat) %*% h_hat
  }
  
  # compute phi2_hat
  phi2_hat <<- vector(length = observation_count)
  for (i in 1:observation_count) {
    phi2_hat[i] <<- t(a_hat) %*% sigma_hat_inverse %*% b_hat - (g_hat_transpose %*% x[i,]) * (t(h_hat) %*% x[i,])
  }
  
  # compute phi3_hat
  optim2 <<- (t(x) %*% x) %*% solve(t(z) %*% x) # similar logic to optim1
  phi3_hat <<- vector(length = observation_count)
  for (i in 1:observation_count) {
    phi3_hat[i] <<- g_hat_transpose %*% (x[i,] * e_hat[i] - (optim2 %*% z[i,] * e_hat[i]) - b_hat)
  }
  
  # compute variance
  variance <<- 0
  for (i in 1:observation_count) {
    variance <<- variance + (phi1_hat[i] + phi2_hat[i] + phi3_hat[i])^2
  }
  
  variance <<- variance / observation_count^2
  
  return(variance)
}

# calculate the variance of the effect of endogeneity in the cubic case
calcEndogeneityEffectVarianceAlternate = function() {
  optim1 <<- solve(t(z_old) %*% x_old) # only compute (t(z) %*% x)^-1 only once, as the value is the same each iteration, to speed up the program
  e_hat <<- vector(length = observation_count)
  for (i in 1:observation_count) {
    e_hat[i] <<- data$budget_share[[i]] - t(x_old[i,]) %*% optim1 %*% t(z_old) %*% data$budget_share; 
  }
  
  # compute b_hat
  b_hat <<- rep(0,regressor_count)
  for (i in 1:observation_count) {
    b_hat <<- b_hat + x[i,] * e_hat[i]
  }
  b_hat <<- b_hat / observation_count
  
  sigma_hat_inverse = solve(sigma_hat)
  
  g_hat_transpose <<- t(a_hat) %*% sigma_hat_inverse # 1x5
  h_hat <<- sigma_hat_inverse %*% b_hat # 5x1
  
  # compute phi1_hat
  phi1_hat <<- vector(length = observation_count)
  for (i in 1:observation_count) {
    phi1_hat[i] <<- t(a[i,] - a_hat) %*% h_hat
  }
  
  # compute phi2_hat
  phi2_hat <<- vector(length = observation_count)
  for (i in 1:observation_count) {
    phi2_hat[i] <<- t(a_hat) %*% sigma_hat_inverse %*% b_hat - (g_hat_transpose %*% x[i,]) * (t(h_hat) %*% x[i,])
  }
  
  # compute phi3_hat
  optim2 <<- (t(x) %*% x_old) %*% solve(t(z_old) %*% x_old) # similar logic to optim1
  phi3_hat <<- vector(length = observation_count)
  for (i in 1:observation_count) {
    phi3_hat[i] <<- g_hat_transpose %*% (x[i,] * e_hat[i] - (optim2 %*% z_old[i,] * e_hat[i]) - b_hat)
  }
  
  # compute variance
  variance <<- 0
  for (i in 1:observation_count) {
    variance <<- variance + (phi1_hat[i] + phi2_hat[i] + phi3_hat[i])^2
  }
  
  variance <<- variance / observation_count^2
  
  return(variance)
}

# computer the predicted quantity function, and return it
# note that the format for the agruments price_powers is slightly different,
# from earlier functions in that income_powers and price_powers take a leading 0
# income_powers is like price_powers, but for income
quantFunc = function(beta_hat, regressors, price_powers, income_powers) {
  if (length(regressors)+1 != length(price_powers) || length(income_powers) != length(price_powers)) {
    print("Lengths Powers do not match in length")
    return(NULL)
  }
  
  ln_y_bar <<- mean(data$ln_income_dollars) # mean of the log of the income
  mean_income_weekly <<- mean(e^data$ln_income_dollars)/52 # correct for the fact that income is yearly

  leftover_means <<- vector(length = length(regressors)+1)
  leftover_means[1] <<- 1
  
  for (i in 1:length(regressors)) {
    leftover_means[i+1] <<- mean(regressors[[i]] / (data$ln_price_dollars ^ price_powers[[i+1]]) / (data$ln_income_dollars ^ income_powers[[i+1]]))
  }
  
  return(function(k) { # k is a vector of prices
    out <<- vector(length = length(k)) # out is a vector of quantities
    
    i <- 1
    for (p in k) {
      out[i] <<- mean_income_weekly / p * exec(paste(beta_hat * leftover_means, " * (", log(p), " ^ ", price_powers, ") * (", ln_y_bar, " ^ ", income_powers, ")", collapse = "+", sep=""))
      i <- i + 1
    }
    
    return(out)
  })
}

### OUTPUT ###

{
print("BEGIN PROGRAM OUTPUT")

set.seed(0) # for the sake of being able to re-produce results

# Linear 
influence_linear = calcInfluenceFunction(list(data$num_drivers_household, data$mean_age, data$ln_price_dollars, data$ln_income_dollars), list(0, 0, 1, 0))
cs_linear = a_hat %*% beta_hat
se_linear = sqrt(var(influence_linear) / observation_count)

func_linear = quantFunc(beta_hat, list(data$num_drivers_household, data$mean_age, data$ln_price_dollars, data$ln_income_dollars), 
                        list(0, 0, 0, 1, 0), 
                        list(0, 0, 0, 0, 1))

curve(func_linear, from=1.14, to=1.46, ylab = "predicted quantity", xlab = "price", main = "linear")
quant_linear = func_linear(e^data$ln_price_dollars)

effect_linear = calcEndogeneityEffect(list(data$num_drivers_household, data$mean_age, data$ln_price_dollars, data$ln_income_dollars), list(0, 0, 1, 0))

effect_linear_variance = calcEndogeneityEffectVariance()

# Quadratic
influence_quadratic = calcInfluenceFunction(list(data$num_drivers_household, data$mean_age,  data$ln_price_dollars, data$ln_income_dollars,
                                                 data$ln_price_dollars^2, data$ln_income_dollars^2, data$ln_price_dollars*data$ln_income_dollars),
                                            list(0,0,1,0,2,0,1))
cs_quadratic = a_hat %*% beta_hat
se_quadratic = sqrt(var(influence_quadratic) / observation_count)

effect_quadratic = calcEndogeneityEffect(list(data$num_drivers_household, data$mean_age,  data$ln_price_dollars, data$ln_income_dollars,
                                              data$ln_price_dollars^2, data$ln_income_dollars^2, data$ln_price_dollars*data$ln_income_dollars),
                                        list(0,0,1,0,2,0,1))

func_quadratic = quantFunc(beta_hat, list(data$num_drivers_household, data$mean_age,  data$ln_price_dollars, data$ln_income_dollars,
                                          data$ln_price_dollars^2, data$ln_income_dollars^2, data$ln_price_dollars*data$ln_income_dollars), 
                           list(0,0,0,1,0,2,0,1), 
                           list(0,0,0,0,1,0,2,1))

curve(func_quadratic, from=1.14, to=1.46, ylab = "predicted quantity", xlab = "price", main = "quadratic")
quant_quadratic = func_quadratic(e^data$ln_price_dollars)

effect_quadratic_variance = calcEndogeneityEffectVariance()

# for use in the cubic case

u_hat_old <- u_hat
x_old <- x
z_old <- z
a_old <- a
a_hat_old <- a_hat

# Cubic
influence_cubic = calcInfluenceFunction(list(data$num_drivers_household, data$mean_age,  data$ln_price_dollars, data$ln_income_dollars,
                                                  data$ln_price_dollars^2, data$ln_income_dollars^2, data$ln_price_dollars*data$ln_income_dollars,
                                                  data$ln_price_dollars^3, data$ln_price_dollars^2*data$ln_income_dollars, 
                                                  data$ln_price_dollars*data$ln_income_dollars^2, data$ln_income_dollars^3),
                                             list(0,0,1,0,2,0,1,3,2,1,0))
cs_cubic = a_hat %*% beta_hat
se_cubic = sqrt(var(influence_cubic) / observation_count)

# this has a singular matrix
#effect_cubic = calcEndogeneityEffect(list(data$num_drivers_household, data$mean_age,  data$ln_price_dollars, data$ln_income_dollars,
#                                         data$ln_price_dollars^2, data$ln_income_dollars^2, data$ln_price_dollars*data$ln_income_dollars,
#                                         data$ln_price_dollars^3, data$ln_price_dollars^2*data$ln_income_dollars, 
#                                         data$ln_price_dollars*data$ln_income_dollars^2, data$ln_income_dollars^3),
#                                    list(0,0,1,0,2,0,1,3,2,1,0))
effect_cubic = a_hat %*% solve(sigma_hat) %*% t(x) %*% (u_hat_old / observation_count)
effect_cubic_variance = calcEndogeneityEffectVarianceAlternate() # if not for singular matrix -> calcEndogeneityEffectVariance()

func_cubic = quantFunc(beta_hat, list(data$num_drivers_household, data$mean_age,  data$ln_price_dollars, data$ln_income_dollars,
                                      data$ln_price_dollars^2, data$ln_income_dollars^2, data$ln_price_dollars*data$ln_income_dollars,
                                      data$ln_price_dollars^3, data$ln_price_dollars^2*data$ln_income_dollars, 
                                      data$ln_price_dollars*data$ln_income_dollars^2, data$ln_income_dollars^3), 
                           list(0,0,0,1,0,2,0,1,3,2,1,0), 
                           list(0,0,0,0,1,0,2,1,0,1,2,3))

curve(func_cubic, from=1.14, to=1.46, ylab = "predicted quantity", xlab = "price", main = "cubic")

quant_cubic = func_cubic(e^data$ln_price_dollars)

print("")
print("Completion notice: ")
print(paste("Linear    Comsumer Surplus Estimate: ", cs_linear, "with standard error", se_linear))
print(paste("Quadratic Comsumer Surplus Estimate: ", cs_quadratic, "with standard error", se_quadratic))
print(paste("Cubic     Comsumer Surplus Estimate: ", cs_cubic, "with standard error", se_cubic))
print(paste("Linear    Endogeneity Effect", effect_linear, " with variance", effect_linear_variance))
print(paste("Quadratic Endogeneity Effect", effect_quadratic, "with variance", effect_quadratic_variance))
print(paste("Cubic     Endogeneity Effect", effect_cubic, "with variance", effect_cubic_variance))

}
